'''common.py: common classes and functions'''
#
# Copyright (C) 2013-2016 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function
import atexit
import codecs
import inspect
import json
import logging
import magic
import os
import re
import shutil
import subprocess
import sys
import tempfile
import types


DEBUGGING = False
UNPACK_DIR = None
RAW_UNPACK_DIR = None
TMP_DIR = None
VALID_SYSCALL = r'^[a-z0-9_]{2,64}$'
# There are quite a few kernel interfaces that can cause problems with
# long profile names. These are outlined in
# https://launchpad.net/bugs/1499544. The big issue is that the audit
# message must fit within PAGE_SIZE (at least 4096 on supported archs),
# so long names could push the audit message to be too big, which would
# result in a denial for that rule (but, only if the rule would've
# allowed it). Giving a hard-error on maxlen since we know that this
# will be a problem. The advisory length is what it is since we know
# that compound labels are sometimes logged and so a snappy system
# running an app in a snappy container or a QA testbed running apps
# under LXC
AA_PROFILE_NAME_MAXLEN = 230  # 245 minus a bit for child profiles
AA_PROFILE_NAME_ADVLEN = 100
# Store enforces this length for snap v2
STORE_PKGNAME_SNAPV2_MAXLEN = 40


def cleanup_unpack():
    global UNPACK_DIR
    if UNPACK_DIR is not None and os.path.isdir(UNPACK_DIR):
        recursive_rm(UNPACK_DIR)
        UNPACK_DIR = None
    global RAW_UNPACK_DIR
    if RAW_UNPACK_DIR is not None and os.path.isdir(RAW_UNPACK_DIR):
        recursive_rm(RAW_UNPACK_DIR)
        RAW_UNPACK_DIR = None
    global TMP_DIR
    if TMP_DIR is not None and os.path.isdir(TMP_DIR):
        recursive_rm(TMP_DIR)
        TMP_DIR = None


atexit.register(cleanup_unpack)


#
# Utility classes
#
class ReviewException(Exception):
    '''This class represents Review exceptions'''
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class Review(object):
    '''Common review class'''
    magic_binary_file_descriptions = [
        'application/x-executable; charset=binary',
        'application/x-sharedlib; charset=binary',
        'application/x-object; charset=binary',
        'application/x-executable',
        'application/x-sharedlib',
        'application/x-object',
    ]

    def __init__(self, fn, review_type, overrides=None):
        self.pkg_filename = fn
        self._check_package_exists()

        self.review_type = review_type
        # TODO: rename as pkg_report
        self.click_report = dict()
        self.stage_report = dict()

        self.result_types = ['info', 'warn', 'error']
        for r in self.result_types:
            self.click_report[r] = dict()
            self.stage_report[r] = dict()

        self.click_report_output = "json"

        global UNPACK_DIR
        if UNPACK_DIR is None:
            UNPACK_DIR = unpack_pkg(fn)
        self.unpack_dir = UNPACK_DIR

        global RAW_UNPACK_DIR
        if RAW_UNPACK_DIR is None:
            RAW_UNPACK_DIR = raw_unpack_pkg(fn)
        self.raw_unpack_dir = RAW_UNPACK_DIR

        self.is_click = False
        self.is_snap1 = False
        self.is_snap2 = False
        self.pkgfmt = {"type": "", "version": ""}

        (self.pkgfmt["type"], pkgver) = detect_package(fn, self.unpack_dir)

        if self._pkgfmt_type() == "snap":
            if pkgver < 2:
                self.is_snap1 = True
                self.pkgfmt["version"] = "15.04"
            else:
                self.is_snap2 = True
                self.pkgfmt["version"] = "16.04"
        elif self._pkgfmt_type() == "click":
            self.pkgfmt["version"] = "0.4"
            self.is_click = True
        else:
            error("Unknown package type: '%s'" % self._pkgfmt_type())

        # Get a list of all unpacked files
        self.pkg_files = []
        self._list_all_files()

        # Setup what is needed to get a list of all unpacked compiled binaries
        self.mime = magic.open(magic.MAGIC_MIME)
        self.mime.load()
        self.pkg_bin_files = []
        # Don't run this here since only cr_lint.py and cr_functional.py need
        # it now
        # self._list_all_compiled_binaries()

        self.overrides = overrides if overrides is not None else {}

        self.override_result_type = None

    def _check_innerpath_executable(self, fn):
        '''Check that the provided path exists and is executable'''
        return os.access(fn, os.X_OK)

    def _extract_statinfo(self, fn):
        '''Extract statinfo from file'''
        try:
            st = os.stat(fn)
        except Exception:
            return None
        return st

    def _extract_file(self, rel):
        '''Extract file'''
        fn = os.path.join(self.unpack_dir, rel)
        if not os.path.isfile(fn):
            error("Could not find '%s'" % rel)
        return open_file_read(fn)

    def _path_join(self, dirname, rest):
        return os.path.join(dirname, rest)

    def _get_sha512sum(self, fn):
        '''Get sha512sum of file'''
        (rc, out) = cmd(['sha512sum', fn])
        if rc != 0:
            return None
        return out.split()[0]

    def _pkgfmt_type(self):
        '''Return the package format type'''
        if "type" not in self.pkgfmt:
            return ""
        return self.pkgfmt["type"]

    def _pkgfmt_version(self):
        '''Return the package format version'''
        if "version" not in self.pkgfmt:
            return ""
        return self.pkgfmt["version"]

    def _check_package_exists(self):
        '''Check that the provided package exists'''
        if not os.path.exists(self.pkg_filename):
            error("Could not find '%s'" % self.pkg_filename)

    def _list_all_files(self):
        '''List all files included in this click package.'''
        for root, dirnames, filenames in os.walk(self.unpack_dir):
            for f in filenames:
                self.pkg_files.append(os.path.join(root, f))

    def _check_if_message_catalog(self, fn):
        '''Check if file is a message catalog (.mo file).'''
        if fn.endswith('.mo'):
            return True
        return False

    def _list_all_compiled_binaries(self):
        '''List all compiled binaries in this click package.'''
        for i in self.pkg_files:
            try:
                res = self.mime.file(i)
            except Exception:  # pragma: nocover
                # workaround for zesty python3-magic
                debug("could not detemine mime type of '%s'" % i)
                continue

            if res in self.magic_binary_file_descriptions and \
               not self._check_if_message_catalog(i) and \
               i not in self.pkg_bin_files:
                self.pkg_bin_files.append(i)

    def _get_check_name(self, name, app='', extra=''):
        name = ':'.join([self.review_type, name])
        if app:
            name += ':' + app
        if extra:
            name += ':' + extra
        return name

    def _verify_pkgversion(self, v):
        '''Verify package name'''
        if not isinstance(v, (str, int, float)):
            return False
        re_valid_version = re.compile(r'^((\d+):)?'              # epoch
                                      '([A-Za-z0-9.+:~-]+?)'     # upstream
                                      '(-([A-Za-z0-9+.~]+))?$')  # debian
        if re_valid_version.match(str(v)):
            return True
        return False

    # click_report[<result_type>][<review_name>] = <result>
    #   result_type: info, warn, error
    #   review_name: name of the check (prefixed with self.review_type)
    #   result: contents of the review
    #   link: url for more information
    #   manual_review: force manual review
    #   override_result_type: prefix results with [<result_type>] and set
    #     result_type to override_result_type
    def _add_result(self, result_type, review_name, result, link=None,
                    manual_review=False, override_result_type=None,
                    stage=False):
        '''Add result to report'''
        if stage:
            report = self.stage_report
        else:
            report = self.click_report

        if result_type not in self.result_types:
            error("Invalid result type '%s'" % result_type)

        prefix = ""
        if override_result_type is not None:
            if override_result_type not in self.result_types:
                error("Invalid override result type '%s'" %
                      override_result_type)
            prefix = "[%s] " % result_type.upper()
            result_type = override_result_type

        if review_name not in report[result_type]:
            # log info about check so it can be collected into the
            # check-names.list file
            # format should be
            # CHECK|<review_type:check_name>|<link>
            msg = 'CHECK|{}|{}'
            name = ':'.join(review_name.split(':')[:2])
            link_text = link if link is not None else ""
            logging.debug(msg.format(name, link_text))
            report[result_type][review_name] = dict()

        report[result_type][review_name].update({
            'text': "%s%s" % (prefix, result),
            'manual_review': manual_review,
        })
        if link is not None:
            report[result_type][review_name]["link"] = link

    def _apply_staged_results(self):
        '''Merge the staged report into the main report'''
        for result_type in self.stage_report:
            if result_type not in self.result_types:
                error("Invalid result type '%s'" % result_type)

            for review_name in self.stage_report[result_type]:
                if review_name not in self.click_report[result_type]:
                    self.click_report[result_type][review_name] = dict()
                for key in self.stage_report[result_type][review_name]:
                    self.click_report[result_type][review_name][key] = \
                        self.stage_report[result_type][review_name][key]
            # reset the staged report
            self.stage_report[result_type] = dict()

    def do_report(self):
        '''Print report'''
        if self.click_report_output == "console":
            # TODO: format better
            import pprint
            pprint.pprint(self.click_report)
        elif self.click_report_output == "json":
            import json
            msg(json.dumps(self.click_report,
                           sort_keys=True,
                           indent=2,
                           separators=(',', ': ')))

        rc = 0
        if len(self.click_report['error']):
            rc = 2
        elif len(self.click_report['warn']):
            rc = 1
        return rc

    def do_checks(self):
        '''Run all methods that start with check_'''
        methodList = [name for name, member in
                      inspect.getmembers(self, inspect.ismethod)
                      if isinstance(member, types.MethodType)]
        for methodname in methodList:
            if not methodname.startswith("check_"):
                continue
            func = getattr(self, methodname)
            func()

    def set_review_type(self, name):
        '''Set review name'''
        self.review_type = name


#
# Utility functions
#

def error(out, exit_code=1, do_exit=True):
    '''Print error message and exit'''
    try:
        print("ERROR: %s" % (out), file=sys.stderr)
    except IOError:
        pass

    if do_exit:
        sys.exit(exit_code)


def warn(out):
    '''Print warning message'''
    try:
        print("WARN: %s" % (out), file=sys.stderr)
    except IOError:
        pass


def msg(out, output=sys.stdout):
    '''Print message'''
    try:
        print("%s" % (out), file=output)
    except IOError:
        pass


def debug(out):
    '''Print debug message'''
    global DEBUGGING
    if DEBUGGING:
        try:
            print("DEBUG: %s" % (out), file=sys.stderr)
        except IOError:
            pass


def cmd(command):
    '''Try to execute the given command.'''
    debug(command)
    try:
        sp = subprocess.Popen(command, stdout=subprocess.PIPE,
                              stderr=subprocess.STDOUT)
    except OSError as ex:
        return [127, str(ex)]

    if sys.version_info[0] >= 3:
        out = sp.communicate()[0].decode('ascii', 'ignore')
    else:
        out = sp.communicate()[0]

    return [sp.returncode, out]


def cmd_pipe(command1, command2):
    '''Try to pipe command1 into command2.'''
    try:
        sp1 = subprocess.Popen(command1, stdout=subprocess.PIPE)
        sp2 = subprocess.Popen(command2, stdin=sp1.stdout)
    except OSError as ex:
        return [127, str(ex)]

    if sys.version_info[0] >= 3:
        out = sp2.communicate()[0].decode('ascii', 'ignore')
    else:
        out = sp2.communicate()[0]

    return [sp2.returncode, out]


def _unpack_cmd(cmd_args, d, dest):
    '''Low level unpack helper'''
    curdir = os.getcwd()
    os.chdir(d)

    (rc, out) = cmd(cmd_args)
    os.chdir(curdir)

    if rc != 0:
        if os.path.isdir(d):
            recursive_rm(d)
        error("unpacking failed with '%d':\n%s" % (rc, out))

    if dest is None:
        dest = d
    else:
        shutil.move(d, dest)

    return dest


def _unpack_snap_squashfs(snap_pkg, dest):
    '''Unpack a squashfs based snap package to dest'''
    d = tempfile.mkdtemp(prefix='review-')
    return _unpack_cmd(['unsquashfs', '-f', '-d', d,
                        os.path.abspath(snap_pkg)], d, dest)


def _unpack_click_deb(pkg, dest):
    d = tempfile.mkdtemp(prefix='review-')
    return _unpack_cmd(['dpkg-deb', '-R',
                        os.path.abspath(pkg), d], d, dest)


def unpack_pkg(fn, dest=None):
    '''Unpack package'''
    if not os.path.isfile(fn):
        error("Could not find '%s'" % fn)
    pkg = fn
    if not pkg.startswith('/'):
        pkg = os.path.abspath(pkg)

    if dest is not None and os.path.exists(dest):
        error("'%s' exists. Aborting." % dest)

    # check if its a squashfs based snap
    if is_squashfs(pkg):
        return _unpack_snap_squashfs(fn, dest)

    return _unpack_click_deb(fn, dest)


def is_squashfs(filename):
    '''Return true if the given filename as a squashfs header'''
    with open(filename, 'rb') as f:
        header = f.read(10)
    return header.startswith(b"hsqs")


def raw_unpack_pkg(fn, dest=None):
    '''Unpack raw package'''
    if not os.path.isfile(fn):
        error("Could not find '%s'" % fn)
    pkg = fn
    if not pkg.startswith('/'):
        pkg = os.path.abspath(pkg)
    # nothing to do for squashfs images
    if is_squashfs(pkg):
        return ""

    if dest is not None and os.path.exists(dest):
        error("'%s' exists. Aborting." % dest)

    d = tempfile.mkdtemp(prefix='review-')

    curdir = os.getcwd()
    os.chdir(d)
    (rc, out) = cmd(['ar', 'x', pkg])
    os.chdir(curdir)

    if rc != 0:
        if os.path.isdir(d):
            recursive_rm(d)
        error("'ar x' failed with '%d':\n%s" % (rc, out))

    if dest is None:
        dest = d
    else:
        shutil.move(d, dest)

    return dest


def create_tempdir():
    '''Create/reuse a temporary directory that is automatically cleaned up'''
    global TMP_DIR
    if TMP_DIR is None:
        TMP_DIR = tempfile.mkdtemp(prefix='review-')
    return TMP_DIR


def open_file_read(path):
    '''Open specified file read-only'''
    try:
        orig = codecs.open(path, 'r', "UTF-8")
    except Exception:
        raise

    return orig


def recursive_rm(dirPath, contents_only=False):
    '''recursively remove directory'''
    try:
        names = os.listdir(dirPath)
    except PermissionError:
        # If directory has weird permissions (eg, 000), just try to remove the
        # directory if we can. If it is non-empty, we'll legitimately fail
        # here. This allows us to remove empty directories with weird
        # permissions.
        os.rmdir(dirPath)
        return

    for name in names:
        path = os.path.join(dirPath, name)
        if os.path.islink(path) or not os.path.isdir(path):
            os.unlink(path)
        else:
            recursive_rm(path)
    if contents_only is False:
        os.rmdir(dirPath)


def run_check(cls):
    if len(sys.argv) < 2:
        error("Must give path to package")

    # extract args
    fn = sys.argv[1]
    if len(sys.argv) > 2:
        overrides = json.loads(sys.argv[2])
    else:
        overrides = None

    review = cls(fn, overrides=overrides)
    review.do_checks()
    rc = review.do_report()
    sys.exit(rc)


def detect_package(fn, dir=None):
    '''Detect what type of package this is'''
    pkgtype = None
    pkgver = None

    if not os.path.isfile(fn):
        error("Could not find '%s'" % fn)

    if dir is None:
        unpack_dir = unpack_pkg(fn)
    else:
        unpack_dir = dir

    if not os.path.isdir(unpack_dir):
        error("Could not find '%s'" % unpack_dir)

    pkg = fn
    if not pkg.startswith('/'):
        pkg = os.path.abspath(pkg)

    # check if its a squashfs based snap
    if is_squashfs(pkg):
        # 16.04+ squashfs snaps
        pkgtype = "snap"
        pkgver = 2
    elif os.path.exists(os.path.join(unpack_dir, "meta/package.yaml")):
        # 15.04 ar-based snaps
        pkgtype = "snap"
        pkgver = 1
    else:
        pkgtype = "click"
        pkgver = 1

    if dir is None and os.path.isdir(unpack_dir):
        recursive_rm(unpack_dir)

    return (pkgtype, pkgver)


def find_external_symlinks(unpack_dir, pkg_files, pkgname):
    '''Check if symlinks in the package go out to the system.'''
    common = r'(-[0-9.]+)?\.so(\.[0-9.]+)?'
    libc6_libs = ['ld-*.so',
                  'libanl',
                  'libBrokenLocale',
                  'libc',
                  'libcidn',
                  'libcrypt',
                  'libdl',
                  'libmemusage',
                  'libm',
                  'libmvec',
                  'libnsl',
                  'libnss_compat',
                  'libnss_dns',
                  'libnss_files',
                  'libnss_hesiod',
                  'libnss_nisplus',
                  'libnss_nis',
                  'libpcprofile',
                  'libpthread',
                  'libresolv',
                  'librt',
                  'libSegFault',
                  'libthread_db',
                  'libutil',
                  ]
    libc6_pats = []
    for lib in libc6_libs:
        libc6_pats.append(re.compile(r'%s%s' % (lib, common)))
    libc6_pats.append(re.compile(r'ld-*.so$'))
    libc6_pats.append(re.compile(r'ld-linux-.*.so\.[0-9.]+$'))
    libc6_pats.append(re.compile(r'ld64.so\.[0-9.]+$'))  # ppc64el

    def _in_patterns(pats, f):
        for pat in pats:
            if pat.search(f):
                return True
        return False

    def _is_external(link, pats, pkgname):
        rp = os.path.realpath(link)
        if not rp.startswith(unpack_dir + "/") and \
                not rp.startswith(os.path.join("/snap", pkgname) + "/") and \
                not rp.startswith(
                    os.path.join("/var/snap", pkgname) + "/") and \
                not _in_patterns(pats, os.path.basename(link)):
            return True
        return False

    external_symlinks = list(filter(lambda link:
                             _is_external(link, libc6_pats, pkgname),
                             pkg_files))

    return [os.path.relpath(i, unpack_dir) for i in external_symlinks]


# check_results(report, expected_counts, expected)
# Verify exact counts of types
#   expected_counts={'info': 1, 'warn': 0, 'error': 0}
#   self.check_results(report, expected_counts)
# Verify counts of warn and error types
#   expected_counts={'info': None, 'warn': 0, 'error': 0}
#   self.check_results(report, expected_counts)
# Verify exact messages:
#   expected = dict()
#   expected['info'] = dict()
#   expected['warn'] = dict()
#   expected['warn']['skeleton_baz'] = "TODO"
#   expected['error'] = dict()
#   self.check_results(r, expected=expected)


def check_results(testobj, report,
                  expected_counts={'info': 1, 'warn': 0, 'error': 0},
                  expected=None):
    if expected is not None:
        for t in expected.keys():
            for r in expected[t]:
                testobj.assertTrue(r in report[t],
                                   "Could not find '%s' (%s) in:\n%s" %
                                   (r, t, json.dumps(report, indent=2)))
                for k in expected[t][r]:
                    testobj.assertTrue(k in report[t][r],
                                       "Could not find '%s' (%s) in:\n%s" %
                                       (k, r, json.dumps(report, indent=2)))
                testobj.assertEqual(expected[t][r][k], report[t][r][k])
    else:
        for k in expected_counts.keys():
            if expected_counts[k] is None:
                continue
            testobj.assertEqual(len(report[k]), expected_counts[k],
                                "(%s not equal)\n%s" %
                                (k, json.dumps(report, indent=2)))



'''cr_common.py: common classes and functions'''
#
# Copyright (C) 2013-2016 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function
from debian.deb822 import Deb822
import glob
import json
import os
import pprint
import re
import yaml


from clickreviews.common import (
    Review,
    ReviewException,
    error,
    open_file_read,
)


#
# Utility classes
#
class ClickReviewException(ReviewException):
    '''This class represents ClickReview exceptions'''


class ClickReview(Review):
    '''This class represents click reviews'''
    # Convenience to break out common types of clicks (eg, app, scope,
    # click service)
    app_allowed_peer_hooks = ["account-application",
                              "account-service",
                              "accounts",
                              "account-provider",
                              "account-qml-plugin",
                              "apparmor",
                              "content-hub",
                              "desktop",
                              "push-helper",
                              "urls",
                              ]
    scope_allowed_peer_hooks = ["account-application",
                                "account-service",
                                "accounts",
                                "apparmor",
                                "scope",
                                ]
    # FIXME: when apparmor-policy is implemented, use this
    service_allowed_peer_hooks = ["apparmor",
                                  "bin-path",  # obsoleted, ignored
                                  "snappy-systemd",  # obsoleted, ignored
                                  ]

    snappy_required = ["name", "version"]
    # optional snappy fields here (may be required by appstore)
    snappy_optional = ["architecture",
                       "architectures",
                       "binaries",
                       "caps",
                       "config",
                       "dtbs",
                       "firmware",
                       "frameworks",
                       "icon",
                       "immutable-config",
                       "initrd",
                       "kernel",
                       "modules",
                       "oem",
                       "gadget",
                       "services",
                       "source",
                       "type",
                       "vendor",  # deprecated, unused
                       ]
    snappy_exe_security = ["caps",
                           "security-template",
                           "security-override",
                           "security-policy"]

    def __init__(self, fn, review_type, peer_hooks=None, overrides=None,
                 peer_hooks_link=None):
        Review.__init__(self, fn, review_type, overrides=overrides)

        # The cr_* scripts only support 15.04 snaps (v1). Use sr_* scripts for
        # 16.04 (v2) or higher
        if not self.is_click and not self.is_snap1:
            return

        if not self.pkg_filename.endswith(".click") and \
                not self.pkg_filename.endswith(".snap"):
            if self.pkg_filename.endswith(".deb"):
                error("filename does not end with '.click', but '.deb' "
                      "instead. See http://askubuntu.com/a/485544/94326 for "
                      "how click packages are different.")
            error("filename does not end with '.click'")

        self.manifest = None
        self.click_pkgname = None
        self.click_version = None
        self.pkg_arch = []
        self.is_snap_oem = False
        self.peer_hooks = peer_hooks
        self.peer_hooks_link = peer_hooks_link

        if self.is_snap1:
            pkg_yaml = self._extract_package_yaml()
            if pkg_yaml:
                try:
                    self.pkg_yaml = yaml.safe_load(pkg_yaml)
                except Exception:
                    error("Could not load package.yaml. "
                          "Is it properly formatted?")
                self._verify_package_yaml_structure()
            else:
                error("Could not load package.yaml.")

            #  default to 'app'
            if 'type' not in self.pkg_yaml:
                self.pkg_yaml['type'] = 'app'

            if 'architectures' in self.pkg_yaml:
                self.pkg_arch = self.pkg_yaml['architectures']
            elif 'architecture' in self.pkg_yaml:
                if isinstance(self.pkg_yaml['architecture'], str):
                    self.pkg_arch = [self.pkg_yaml['architecture']]
                elif isinstance(self.pkg_yaml['architecture'], list):
                    self.pkg_arch = self.pkg_yaml['architecture']
                else:
                    error("Could not load package.yaml: "
                          "invalid 'architecture'")
            else:
                self.pkg_arch = ['all']

            if 'type' in self.pkg_yaml and self.pkg_yaml['type'] == 'oem':
                self.is_snap_oem = True

        if self.is_click or self.is_snap1:
            # Get some basic information from the control file
            control_file = self._extract_control_file()
            tmp = list(Deb822.iter_paragraphs(control_file))
            if len(tmp) != 1:
                error("malformed control file: too many paragraphs")
            control = tmp[0]
            self.click_pkgname = control['Package']
            self.click_version = control['Version']
            if self.is_click:
                if control['Architecture'] not in self.pkg_arch:
                    self.pkg_arch.append(control['Architecture'])
                self.pkgfmt["version"] = str(control['Click-Version'])

            # Parse and store the manifest
            manifest_json = self._extract_manifest_file()
            try:
                self.manifest = json.load(manifest_json)
            except Exception:
                error("Could not load manifest file. "
                      "Is it properly formatted?")
            self._verify_manifest_structure()

            self.valid_frameworks = self._extract_click_frameworks()

    def _extract_click_frameworks(self):
        '''Extract installed click frameworks'''
        # TODO: update to use libclick API when available
        valid_frameworks = []
        frameworks = sorted(
            glob.glob("/usr/share/click/frameworks/*.framework"))
        if len(frameworks) == 0:
            valid_frameworks.append('ubuntu-sdk-13.10')
        else:
            for f in frameworks:
                valid_frameworks.append(os.path.basename(
                                        os.path.splitext(f)[0]))
        return valid_frameworks

    def _extract_manifest_file(self):
        '''Extract and read the manifest file'''
        m = os.path.join(self.unpack_dir, "DEBIAN/manifest")
        if not os.path.isfile(m):
            error("Could not find manifest file")
        return open_file_read(m)

    def _extract_package_yaml(self):
        '''Extract and read the snappy 15.04 package.yaml'''
        y = os.path.join(self.unpack_dir, "meta/package.yaml")
        if not os.path.isfile(y):
            return None  # snappy packaging is still optional
        return open_file_read(y)

    def _extract_hashes_yaml(self):
        '''Extract and read the snappy hashes.yaml'''
        y = os.path.join(self.unpack_dir, "DEBIAN/hashes.yaml")
        return open_file_read(y)

    def _extract_control_file(self):
        '''Extract '''
        fh = open_file_read(os.path.join(self.unpack_dir, "DEBIAN/control"))
        return fh.readlines()

    def _verify_manifest_structure(self):
        '''Verify manifest has the expected structure'''
        # lp:click doc/file-format.rst
        mp = pprint.pformat(self.manifest)
        if not isinstance(self.manifest, dict):
            error("manifest malformed:\n%s" % self.manifest)

        required = ["name", "version", "framework"]  # click required
        for f in required:
            if f not in self.manifest:
                error("could not find required '%s' in manifest:\n%s" % (f,
                                                                         mp))
            elif not isinstance(self.manifest[f], str):
                error("manifest malformed: '%s' is not str:\n%s" % (f, mp))

        # optional click fields here (may be required by appstore)
        # http://click.readthedocs.org/en/latest/file-format.html
        optional = ["title", "description", "maintainer", "architecture",
                    "installed-size", "icon"]

        # https://developer.ubuntu.com/snappy/guides/packaging-format-apps/
        snappy_optional = ["ports", "source", "type"]

        for f in optional:
            if f in self.manifest:
                if f != "architecture" and \
                   not isinstance(self.manifest[f], str):
                    error("manifest malformed: '%s' is not str:\n%s" % (f, mp))
                elif f == "architecture" and not \
                    (isinstance(self.manifest[f], str) or
                     isinstance(self.manifest[f], list)):
                    error("manifest malformed: '%s' is not str or list:\n%s" %
                          (f, mp))

        # FIXME: this is kinda gross but the best we can do while we are trying
        # to support clicks and native snaps
        if 'type' in self.manifest and self.manifest['type'] == 'oem':
            if 'hooks' in self.manifest:
                error("'hooks' present in manifest with type 'oem'")
            # mock up something for other tests
            self.manifest['hooks'] = {'oem': {'reviewtools': True}}

        # Not required by click, but required by appstore. 'hooks' is assumed
        # to be present in other checks
        if 'hooks' not in self.manifest:
            error("could not find required 'hooks' in manifest:\n%s" % mp)
        if not isinstance(self.manifest['hooks'], dict):
            error("manifest malformed: 'hooks' is not dict:\n%s" % mp)
        # 'hooks' is assumed to be present and non-empty in other checks
        if len(self.manifest['hooks']) < 1:
            error("manifest malformed: 'hooks' is empty:\n%s" % mp)
        for app in self.manifest['hooks']:
            if not isinstance(self.manifest['hooks'][app], dict):
                error("manifest malformed: hooks/%s is not dict:\n%s" % (app,
                                                                         mp))
            # let cr_lint.py handle required hooks
            if len(self.manifest['hooks'][app]) < 1:
                error("manifest malformed: hooks/%s is empty:\n%s" % (app, mp))

        for k in sorted(self.manifest):
            if k not in required + optional + snappy_optional + ['hooks']:
                # click supports local extensions via 'x-...', ignore those
                # here but report in lint
                if k.startswith('x-'):
                    continue
                error("manifest malformed: unsupported field '%s':\n%s" % (k,
                                                                           mp))

    def _verify_package_yaml_structure(self):
        '''Verify package.yaml has the expected structure'''
        # https://developer.ubuntu.com/en/snappy/guides/packaging-format-apps/
        # lp:click doc/file-format.rst
        yp = yaml.dump(self.pkg_yaml, default_flow_style=False, indent=4)
        if not isinstance(self.pkg_yaml, dict):
            error("package yaml malformed:\n%s" % self.pkg_yaml)

        for f in self.snappy_required:
            if f not in self.pkg_yaml:
                error("could not find required '%s' in package.yaml:\n%s" %
                      (f, yp))
            elif f in ['name', 'version']:
                # make sure this is a string for other tests since
                # yaml.safe_load may make it an int, float or str
                self.pkg_yaml[f] = str(self.pkg_yaml[f])

        for f in self.snappy_optional:
            if f in self.pkg_yaml:
                if f in ["architecture", "frameworks"] and not \
                    (isinstance(self.pkg_yaml[f], str) or
                     isinstance(self.pkg_yaml[f], list)):
                    error("yaml malformed: '%s' is not str or list:\n%s" %
                          (f, yp))
                elif f in ["binaries", "services"] and not \
                        isinstance(self.pkg_yaml[f], list):
                    error("yaml malformed: '%s' is not list:\n%s" % (f, yp))
                elif f in ["icon", "source", "type", "vendor"] and not \
                        isinstance(self.pkg_yaml[f], str):
                    error("yaml malformed: '%s' is not str:\n%s" % (f, yp))

    def _verify_peer_hooks(self, my_hook):
        '''Compare manifest for required and allowed hooks'''
        d = dict()
        if self.peer_hooks is None:
            return d

        for app in self.manifest["hooks"]:
            if my_hook not in self.manifest["hooks"][app]:
                continue
            for h in self.peer_hooks[my_hook]['required']:
                if h == my_hook:
                    continue
                if h not in self.manifest["hooks"][app]:
                    # Treat these as equivalent for satisfying peer hooks
                    if h == 'apparmor' and \
                       'apparmor-profile' in self.manifest["hooks"][app]:
                        continue

                    if 'missing' not in d:
                        d['missing'] = dict()
                    if app not in d['missing']:
                        d['missing'][app] = []
                    d['missing'][app].append(h)
            for h in self.manifest["hooks"][app]:
                if h == my_hook:
                    continue
                if h not in self.peer_hooks[my_hook]['allowed']:
                    # 'apparmor-profile' is allowed when 'apparmor' is, but
                    # they may not be used together
                    if h == 'apparmor-profile':
                        if 'apparmor' in self.peer_hooks[my_hook]['allowed'] \
                           and 'apparmor' not in self.manifest["hooks"][app]:
                            continue

                    if 'disallowed' not in d:
                        d['disallowed'] = dict()
                    if app not in d['disallowed']:
                        d['disallowed'][app] = []
                    d['disallowed'][app].append(h)

        return d

    def _verify_pkgname(self, n):
        '''Verify package name'''
        if self.is_snap1:
            # snaps can't have '.' in the name
            pat = re.compile(r'^[a-z0-9][a-z0-9+-]+$')
        else:
            pat = re.compile(r'^[a-z0-9][a-z0-9+.-]+$')
        if pat.search(n):
            return True
        return False

    def _verify_maintainer(self, m):
        '''Verify maintainer email'''
        #  Simple regex as used by python3-debian. If we wanted to be more
        #  thorough we could use email_re from django.core.validators
        if re.search(r"^(.*)\s+<(.*@.*)>$", m):
            return True
        return False

    def _create_dict(self, lst, topkey='name'):
        '''Converts list of dicts into dict[topkey][<the rest>]. Useful for
           conversions from yaml list to json dict'''
        d = dict()
        for entry in lst:
            if topkey not in entry:
                error("required field '%s' not present: %s" % (topkey, entry))
            name = entry[topkey]
            d[name] = dict()
            for key in entry:
                if key == topkey:
                    continue
                d[name][key] = entry[key]
        return d

    def _get_policy_versions(self, vendor):
        '''Get the supported AppArmor policy versions'''
        if not self.aa_policy:
            return None

        if vendor not in self.aa_policy:
            error("Could not find vendor '%s'" % vendor, do_exit=False)
            return None

        supported_policy_versions = []
        for i in self.aa_policy[vendor].keys():
            supported_policy_versions.append("%.1f" % float(i))

        return sorted(supported_policy_versions)

    def _get_templates(self, vendor, version, aa_type="all"):
        '''Get templates by type'''
        if not self.aa_policy:
            return None

        templates = []
        if aa_type == "all":
            for k in self.aa_policy[vendor][version]['templates'].keys():
                templates += self.aa_policy[vendor][version]['templates'][k]
        else:
            templates = self.aa_policy[vendor][version]['templates'][aa_type]

        return sorted(templates)

    def _has_policy_version(self, vendor, version):
        '''Determine if has specified policy version'''
        if not self.aa_policy:
            return None

        if vendor not in self.aa_policy:
            error("Could not find vendor '%s'" % vendor, do_exit=False)
            return False

        if str(version) not in self.aa_policy[vendor]:
            return False
        return True

    def _get_policy_groups(self, vendor, version, aa_type="all"):
        '''Get policy groups by type'''
        if not self.aa_policy:
            return None

        groups = []
        if vendor not in self.aa_policy:
            error("Could not find vendor '%s'" % vendor, do_exit=False)
            return groups

        if not self._has_policy_version(vendor, version):
            error("Could not find version '%s'" % version, do_exit=False)
            return groups

        v = str(version)
        if aa_type == "all":
            for k in self.aa_policy[vendor][v]['policy_groups'].keys():
                groups += self.aa_policy[vendor][v]['policy_groups'][k]
        else:
            groups = self.aa_policy[vendor][v]['policy_groups'][aa_type]

        return sorted(groups)

    def _get_policy_group_type(self, vendor, version, policy_group):
        '''Return policy group type'''
        if not self.aa_policy:
            return None

        policy_groups = self.aa_policy[vendor][version]['policy_groups']
        for t in policy_groups:
            if policy_group in policy_groups[t]:
                return t
        return None

    def _get_template_type(self, vendor, version, template):
        '''Return template type'''
        if not self.aa_policy:
            return None

        for t in self.aa_policy[vendor][version]['templates']:
            if template in self.aa_policy[vendor][version]['templates'][t]:
                return t

    def check_peer_hooks(self, hooks_sublist=[]):
        '''Check if peer hooks are valid'''
        # Nothing to verify
        if not hasattr(self, 'peer_hooks') or self.peer_hooks is None:
            return

        for hook in self.peer_hooks:
            if len(hooks_sublist) > 0 and hook not in hooks_sublist:
                continue
            d = self._verify_peer_hooks(hook)
            t = 'info'
            n = self._get_check_name("peer_hooks_required", extra=hook)
            s = "OK"

            if 'missing' in d and len(d['missing'].keys()) > 0:
                t = 'error'
                for app in d['missing']:
                    s = "Missing required hooks for '%s': %s" % (
                        app, ", ".join(d['missing'][app]))
                    self._add_result(t, n, s, manual_review=True,
                                     link=self.peer_hooks_link)
            else:
                self._add_result(t, n, s)

            t = 'info'
            n = self._get_check_name("peer_hooks_disallowed", extra=hook)
            s = "OK"

            if 'disallowed' in d and len(d['disallowed'].keys()) > 0:
                t = 'error'
                for app in d['disallowed']:
                    s = "Disallowed with %s (%s): %s" % (
                        hook, app, ", ".join(d['disallowed'][app]))
                    self._add_result(t, n, s, manual_review=True,
                                     link=self.peer_hooks_link)
            else:
                self._add_result(t, n, s)



'''cr_scope.py: click scope'''
#
# Copyright (C) 2014-2015 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function

from clickreviews.cr_common import ClickReview, error
import codecs
import configparser
import os
import re

# Please refer to the config file documentation at:
# https://github.com/ubports/unity-scopes-api/blob/xenial/CONFIGFILES

KNOWN_SECTIONS = set(["ScopeConfig", "Appearance"])


class ClickReviewScope(ClickReview):
    '''This class represents click lint reviews'''
    def __init__(self, fn, overrides=None):
        peer_hooks = dict()
        my_hook = 'scope'
        peer_hooks[my_hook] = dict()
        peer_hooks[my_hook]['allowed'] = ClickReview.scope_allowed_peer_hooks
        peer_hooks[my_hook]['required'] = ['apparmor']

        ClickReview.__init__(self, fn, "scope", peer_hooks=peer_hooks,
                             overrides=overrides)

        if not self.is_click and not self.is_snap1:
            return

        self.scopes = dict()

        if self.manifest is None:
            return

        for app in self.manifest['hooks']:
            if 'scope' not in self.manifest['hooks'][app]:
                # msg("Skipped missing scope hook for '%s'" % app)
                continue
            if not isinstance(self.manifest['hooks'][app]['scope'], str):
                error("manifest malformed: hooks/%s/scope is not str" % app)
            self.scopes[app] = self._extract_scopes(app)

    def _extract_scopes(self, app):
        '''Get scopes'''
        d = dict()

        s = self.manifest['hooks'][app]['scope']
        fn = os.path.join(self.unpack_dir, s)

        bn = os.path.basename(fn)
        if not os.path.exists(fn):
            error("Could not find '%s'" % bn)
        elif not os.path.isdir(fn):
            error("'%s' is not a directory" % bn)

        ini_fn = os.path.join(fn, "%s_%s.ini" % (self.manifest['name'], app))
        ini_fn_bn = os.path.relpath(ini_fn, self.unpack_dir)
        if not os.path.exists(ini_fn):
            error("Could not find scope INI file '%s'" % ini_fn_bn)
        try:
            d["scope_config"] = configparser.ConfigParser()
            d["scope_config"].read_file(codecs.open(ini_fn, "r", "utf8"))
        except Exception as e:
            error("scope config unparseable: %s (%s)" % (ini_fn_bn, str(e)))

        d["dir"] = fn
        d["dir_rel"] = bn
        d["ini_file"] = ini_fn
        d["ini_file_rel"] = ini_fn_bn

        return d

    def check_scope_ini(self):
        '''Check scope .ini file'''
        if not self.is_click and not self.is_snap1:
            return

        for app in sorted(self.scopes.keys()):
            t = 'info'
            n = self._get_check_name('ini_scope_section', app=app)
            s = "OK"

            sections = set(self.scopes[app]["scope_config"].sections())
            unknown_sections = sections.difference(KNOWN_SECTIONS)

            if unknown_sections:
                t = 'error'
                s = "'%s' has unknown sections: %s" % (
                    self.scopes[app]["ini_file_rel"],
                    ", ".join(unknown_sections))
            elif "ScopeConfig" not in sections:
                t = 'error'
                s = "Could not find 'ScopeConfig' in '%s'" % (
                    self.scopes[app]["ini_file_rel"])
                self._add_result(t, n, s)
                continue
            self._add_result(t, n, s)

            # Make these all lower case for easier comparisons
            required = ['author',
                        'description',
                        'displayname']
            optional = ['art',
                        'childscopes',
                        'hotkey',
                        'icon',
                        'idletimeout',
                        'invisible',
                        'keywords',
                        'locationdataneeded',
                        'resultsttltype',
                        'scoperunner',
                        'searchhint']
            translated = ['description',
                          'displayname',
                          'searchhint']
            internal = ['debugmode']

            missing = []
            t = 'info'
            n = self._get_check_name('ini_scope_required_fields', app=app)
            s = "OK"
            for r in required:
                if r not in self.scopes[app]["scope_config"]['ScopeConfig']:
                    missing.append(r)
            if len(missing) == 1:
                t = 'error'
                s = "Missing required field in '%s': %s" % (
                    self.scopes[app]["ini_file_rel"],
                    missing[0])
            elif len(missing) > 1:
                t = 'error'
                s = "Missing required fields in '%s': %s" % (
                    self.scopes[app]["ini_file_rel"],
                    ", ".join(missing))
            self._add_result(t, n, s)

            t = 'info'
            n = self._get_check_name('ini_scope_unknown_fields', app=app)
            s = 'OK'
            unknown = []
            for i in self.scopes[app]["scope_config"]['ScopeConfig'].keys():
                f = i.lower()
                if f not in required and f not in optional and \
                    f not in internal and \
                    (f.split("[")[0] not in translated or not
                        re.search(r'.*\[[a-z]{2,3}(_[a-z]{2,3})?\]$', f)):
                    unknown.append(f)

            if len(unknown) == 1:
                t = 'warn'
                s = "Unknown field in '%s': %s" % (
                    self.scopes[app]["ini_file_rel"],
                    unknown[0])
            elif len(unknown) > 1:
                t = 'warn'
                s = "Unknown fields in '%s': %s" % (
                    self.scopes[app]["ini_file_rel"],
                    ", ".join(unknown))
            self._add_result(t, n, s)

            t = 'info'
            n = self._get_check_name('ini_scope_internal_fields', app=app)
            s = "OK"
            forbidden = []
            for r in internal:
                if r in self.scopes[app]["scope_config"]['ScopeConfig']:
                    forbidden.append(r)
            if len(forbidden) == 1:
                t = 'error'
                s = "Forbidden field in '%s': %s" % (
                    self.scopes[app]["ini_file_rel"],
                    forbidden[0])
            elif len(forbidden) > 1:
                t = 'error'
                s = "Forbidden fields in '%s': %s" % (
                    self.scopes[app]["ini_file_rel"],
                    ", ".join(forbidden))
            self._add_result(t, n, s)



import clickreviews
import imp
import inspect
import os
import pkgutil

IRRELEVANT_MODULES = ['cr_common', 'cr_tests', 'cr_skeleton',
                      'sr_common', 'sr_tests', 'sr_skeleton',
                      'common']


def narrow_down_modules(modules):
    '''
    Get a list of file names or module names and filter out
    the ones we know are irrelevant.
    '''
    relevant_modules = []
    for module in modules:
        module_name = os.path.basename(module).replace('.py', '')
        if module_name not in IRRELEVANT_MODULES and \
                (module_name.startswith('cr_') or
                 module_name.startswith('sr_')):
            relevant_modules += [module]
    return relevant_modules


def get_modules():
    '''
    Here we have a look at all the modules in the
    clickreviews package and filter out a few which
    are not relevant.

    Basically we look at all the ones which are
    derived from [cs]r_common, where we can later on
    instantiate a *Review* object and run the
    necessary checks.
    '''

    all_modules = [name for _, name, _ in
                   pkgutil.iter_modules(clickreviews.__path__)]
    return narrow_down_modules(all_modules)


def find_main_class(module_name):
    '''
    This function will find the Click*Review class in
    the specified module.
    '''
    module = imp.load_source(module_name,
                             '%s/%s.py' % (clickreviews.__path__[0],
                                           module_name))

    classes = inspect.getmembers(module, inspect.isclass)

    def find_test_class(a):
        return (a[0].startswith('Click') or a[0].startswith('Snap')) and \
            not a[0].endswith('Exception') and \
            a[1].__module__ == module_name
    test_class = list(filter(find_test_class, classes))
    if not test_class:
        return None
    init_object = getattr(module, test_class[0][0])
    return init_object


def init_main_class(module_name, click_file, overrides=None):
    '''
    This function will instantiate the main Click*Review
    class of a given module and instantiate it with the
    location of the .click file we want to inspect.
    '''

    init_object = find_main_class(module_name)
    if not init_object:
        return None
    try:
        ob = init_object(click_file, overrides)
    except TypeError as e:
        print('Could not init %s: %s' % (init_object, str(e)))
        raise
    return ob



'''cr_functional.py: click functional'''
#
# Copyright (C) 2013-2015 Canonical Ltd.
# Copyright (C) 2021 UBports Foundation
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function
import binascii
import os
import re
import subprocess

from apt import apt_pkg

from clickreviews.cr_common import ClickReview, open_file_read

# TODO: for QML apps, see if i18n.domain('%s') matches X-Ubuntu-Gettext-Domain
#       compiled apps can use organizationName to match X-Ubuntu-Gettext-Domain

# QML Import versions aligned with Ubuntu Touch framework releases
# Please keep this updated when Qt is upgraded and sorted alphabetically
QT_IMP_FRAMEWORKS = {
    'Qt.labs.folderlistmodel': [
        ('2.1', 'ubuntu-sdk-16.04'),
        ('2.12', 'ubuntu-sdk-16.04.5'),
    ],
    'Qt.labs.platform': [
        ('1.0', 'ubuntu-sdk-16.04'),
    ],
    'Qt.labs.qmlmodels': [
        ('2.12', 'ubuntu-sdk-16.04.5'),
    ],
    'Qt.labs.settings': [
        ('1.0', 'ubuntu-sdk-16.04'),
    ],
    'QtBlueTooth': [
        ('5.9', 'ubuntu-sdk-16.04'),
        ('5.12', 'ubuntu-sdk-16.04.5'),
    ],
    'QtCharts': [
        ('2.2', 'ubuntu-sdk-16.04.3'),
        ('2.3', 'ubuntu-sdk-16.04.5'),
    ],
    'QtGraphicalEffects': [
        ('1.0', 'ubuntu-sdk-16.04'),
        ('1.12', 'ubuntu-sdk-16.04.5'),
    ],
    'QtLocation': [
        ('5.9', 'ubuntu-sdk-16.04'),
        ('5.12', 'ubuntu-sdk-16.04.5'),
    ],
    'QtMultimedia': [
        ('5.9', 'ubuntu-sdk-16.04'),
        ('5.12', 'ubuntu-sdk-16.04.5'),
    ],
    'QtNfc': [
        ('5.12', 'ubuntu-sdk-16.04.6'),
    ],
    'QtPositioning': [
        ('5.9', 'ubuntu-sdk-16.04'),
        ('5.12', 'ubuntu-sdk-16.04.5'),
    ],
    'QtQml': [
        ('2.3', 'ubuntu-sdk-16.04'),
        ('2.12', 'ubuntu-sdk-16.04.5'),
    ],
    'QtQml.Models': [
        ('2.3', 'ubuntu-sdk-16.04'),
        ('2.12', 'ubuntu-sdk-16.04.5'),
    ],
    'QtQml.StateMachine': [
        ('1.0', 'ubuntu-sdk-16.04'),
        ('1.12', 'ubuntu-sdk-16.04.5'),
    ],
    'QtQuick': [
        ('2.9', 'ubuntu-sdk-16.04'),
        ('2.12', 'ubuntu-sdk-16.04.5'),
    ],
    'QtQuick.Controls': [
        ('2.2', 'ubuntu-sdk-16.04'),
        ('2.12', 'ubuntu-sdk-16.04.5'),
    ],
    'QtQuick.Controls.impl': [
        ('2.2', 'ubuntu-sdk-16.04'),
        ('2.12', 'ubuntu-sdk-16.04.5'),
    ],
    'QtQuick.Controls.Fusion': [
        ('2.2', 'ubuntu-sdk-16.04'),
        ('2.12', 'ubuntu-sdk-16.04.5'),
    ],
    'QtQuick.Controls.Fusion.impl': [
        ('2.2', 'ubuntu-sdk-16.04'),
        ('2.12', 'ubuntu-sdk-16.04.5'),
    ],
    'QtQuick.Controls.Imagine': [
        ('2.2', 'ubuntu-sdk-16.04'),
        ('2.12', 'ubuntu-sdk-16.04.5'),
    ],
    'QtQuick.Controls.Imagine.impl': [
        ('2.2', 'ubuntu-sdk-16.04'),
        ('2.12', 'ubuntu-sdk-16.04.5'),
    ],
    'QtQuick.Controls.Material': [
        ('2.2', 'ubuntu-sdk-16.04'),
        ('2.12', 'ubuntu-sdk-16.04.5'),
    ],
    'QtQuick.Controls.Material.impl': [
        ('2.2', 'ubuntu-sdk-16.04'),
        ('2.12', 'ubuntu-sdk-16.04.5'),
    ],
    'QtQuick.Controls.Universal': [
        ('2.2', 'ubuntu-sdk-16.04'),
        ('2.12', 'ubuntu-sdk-16.04.5'),
    ],
    'QtQuick.Controls.Universal.impl': [
        ('2.2', 'ubuntu-sdk-16.04'),
        ('2.12', 'ubuntu-sdk-16.04.5'),
    ],
    'QtQuick.Dialogs': [
        ('1.2', 'ubuntu-sdk-16.04'),
        ('1.3', 'ubuntu-sdk-16.04.5'),
    ],
    'QtQuick.Layouts': [
        ('1.3', 'ubuntu-sdk-16.04'),
        ('1.12', 'ubuntu-sdk-16.04.5'),
    ],
    'QtQuick.LocalStorage': [
        ('2.0', 'ubuntu-sdk-16.04'),
        ('2.12', 'ubuntu-sdk-16.04.5'),
    ],
    'QtQuick.Particles': [
        ('2.0', 'ubuntu-sdk-16.04'),
        ('2.12', 'ubuntu-sdk-16.04.5'),
    ],
    'QtQuick.Shapes': [
        ('1.12', 'ubuntu-sdk-16.04.5'),
    ],
    'QtQuick.Templates': [
        ('2.2', 'ubuntu-sdk-16.04'),
        ('2.12', 'ubuntu-sdk-16.04.5'),
    ],
    'QtQuick.Window': [
        ('2.2', 'ubuntu-sdk-16.04'),
        ('2.12', 'ubuntu-sdk-16.04.5'),
    ],
    'QtQuick.XmlListModel': [
        ('2.0', 'ubuntu-sdk-16.04'),
        ('2.12', 'ubuntu-sdk-16.04.5'),
    ],
    'QtSensors': [
        ('5.9', 'ubuntu-sdk-16.04'),
        ('5.11', 'ubuntu-sdk-16.04.5'),
    ],
    'QtTest': [
        ('1.2', 'ubuntu-sdk-16.04'),
        ('1.12', 'ubuntu-sdk-16.04.5'),
    ],
    'QtWebChannel': [
        ('1.0', 'ubuntu-sdk-16.04.5'),
    ],
    'QtWebEngine': [
        ('1.5', 'ubuntu-sdk-16.04'),
        ('1.7', 'ubuntu-sdk-16.04.1'),
        ('1.10', 'ubuntu-sdk-16.04.4'),
    ],
    'QtWebSockets': [
        ('1.1', 'ubuntu-sdk-16.04'),
    ],
    'QtWebView': [
        ('1.1', 'ubuntu-sdk-16.04.7'),
    ],
    'QtSystemInfo': [
        ('5.5', 'ubuntu-sdk-16.04'),
    ],

    # Additional modules not from Qt directly
    'QtQuick.Controls.Suru': [
        ('2.2', 'ubuntu-sdk-16.04'),
    ],
}

# A dictionary of Qt release series coupled with Ubuntu Touch framework
QT_MIN_FRAMEWORKS = {
    '5.9': 'ubuntu-sdk-16.04',
    '5.12': 'ubuntu-sdk-16.04.5',
}


class ClickReviewFunctional(ClickReview):
    '''This class represents click lint reviews'''
    def __init__(self, fn, overrides=None):
        ClickReview.__init__(self, fn, "functional", overrides=overrides)
        if not self.is_click and not self.is_snap1:
            return

        self.qml_files = []
        for i in self.pkg_files:
            if i.endswith(".qml"):
                self.qml_files.append(i)

        self._list_all_compiled_binaries()

    def check_applicationName(self):
        '''Check applicationName matches click manifest'''
        if not self.is_click and not self.is_snap1:
            return

        if self.manifest is None:
            return

        t = 'info'
        n = self._get_check_name('qml_applicationName_matches_manifest')
        s = "OK"
        link = None

        # find file with MainView in the QML
        mv = r'\s*MainView\s*(\s+{)?'
        pat_mv = re.compile(r'\n%s' % mv)
        qmls = dict()

        for i in self.qml_files:
            qml = open_file_read(i).read()
            if pat_mv.search(qml):
                qmls[i] = qml

        # LP: #1256841 - QML apps with C++ using QSettings shouldn't
        # typically set applicationName in the QML
        for i in self.pkg_bin_files:
            f = open(i, 'rb')
            data = str(binascii.b2a_qp(f.read()))
            f.close()
            if 'QSettings' in data:
                s = "OK (binary uses QSettings)"
                self._add_result(t, n, s)
                return

        if len(self.qml_files) == 0:
            s = "OK (not QML)"
            self._add_result(t, n, s)
            return
        elif len(qmls) == 0:
            s = "SKIP: could not find MainView in QML files"
            self._add_result(t, n, s)
            return

        pat_mvl = re.compile(r'^%s' % mv)
        pat_appname = re.compile(r'^\s*applicationName\s*:\s*["\']')

        ok = False
        appnames = dict()
        for k in qmls.keys():
            in_mainview = False
            for line in qmls[k].splitlines():
                if in_mainview and pat_appname.search(line):
                    appname = line.split(':', 1)[1].strip('"\' \t\n\r\f\v;')
                    appnames[os.path.relpath(k, self.unpack_dir)] = appname
                    if appname == self.click_pkgname:
                        ok = True
                        break
                elif pat_mvl.search(line):
                    in_mainview = True
                if ok:
                    break

        if len(appnames) == 0 or not ok:
            if len(self.pkg_bin_files) == 0:
                t = "warn"
                link = ('http://askubuntu.com/questions/417371/'
                        'what-does-functional-qml-applicationname-matches-'
                        'manifest-mean/417372')

            if len(appnames) == 0:
                s = "could not find applicationName in: %s" % \
                    ", ".join(sorted(list(map(
                        lambda x: os.path.relpath(x, self.unpack_dir), qmls))))
            else:  # not ok
                s = "click manifest name '%s' not found in: " % \
                    self.click_pkgname + "%s" % \
                    ", ".join(sorted(list(map(
                        lambda x: "%s ('%s')" % (x, appnames[x]), appnames))))

            if len(self.pkg_bin_files) == 0:
                s += ". Application may not work properly when confined."
            else:
                s += ". May be ok (detected as compiled application)."

        self._add_result(t, n, s, link)

    def check_qtwebkit(self):
        '''Check that QML applications don't use QtWebKit'''
        if not self.is_click and not self.is_snap1:
            return

        t = 'info'
        n = self._get_check_name('qml_application_uses_QtWebKit')
        s = "OK"
        link = None

        qmls = []
        pat_mv = re.compile(r'\n\s*import\s+QtWebKit')
        for i in self.qml_files:
            qml = open_file_read(i).read()
            if pat_mv.search(qml):
                qmls.append(os.path.relpath(i, self.unpack_dir))

        if len(qmls) > 0:
            t = 'warn'
            s = "Found files that use unsupported QtWebKit (should use " + \
                "UbuntuWebview (Ubuntu.Components.Extras.Browser >= " + \
                "0.2) or Oxide instead): %s" % " ,".join(qmls)
            link = ("http://askubuntu.com/questions/417342/what-does-"
                    "functional-qml-application-uses-qtwebkit-mean/417343")

        self._add_result(t, n, s, link)

        t = 'info'
        n = self._get_check_name('qml_application_uses_UbuntuWebView_0.2')
        s = "OK"
        link = None

        if self.manifest is not None and \
                self.manifest['framework'] == "ubuntu-sdk-13.10":
            s = "SKIPPED (Oxide not available in ubuntu-sdk-13.10)"
        else:
            qmls = []
            pat_mv = re.compile(r'\n\s*import\s+Ubuntu\.Components\.Extras\.'
                                r'Browser\s+0\.1\s*\n')
            for i in self.qml_files:
                qml = open_file_read(i).read()
                if pat_mv.search(qml):
                    qmls.append(os.path.relpath(i, self.unpack_dir))

            if len(qmls) > 0:
                t = 'warn'
                s = "Found files that use unsupported QtWebKit via " + \
                    "'Ubuntu.Components.Extras.Browser 0.1' (should use " + \
                    "Ubuntu.Components.Extras.Browser >= 0.2 or " + \
                    "Oxide instead): %s" % " ,".join(qmls)
                link = ("http://askubuntu.com/questions/417342/what-does-"
                        "functional-qml-application-uses-qtwebkit-mean/417343")

        self._add_result(t, n, s, link)

    def check_friends(self):
        '''Check that QML applications don't use deprecated Friends API'''
        if not self.is_click and not self.is_snap1:
            return

        t = 'info'
        n = self._get_check_name('qml_application_uses_friends')
        s = "OK"
        link = None

        qmls = []
        pat_mv = re.compile(r'\n\s*import\s+Friends')
        for i in self.qml_files:
            qml = open_file_read(i).read()
            if pat_mv.search(qml):
                qmls.append(os.path.relpath(i, self.unpack_dir))

        if len(qmls) > 0:
            t = 'error'
            s = "Found files that use deprecated Friends API: %s" % \
                " ,".join(qmls)
            link = ("http://askubuntu.com/questions/497551/what-does-"
                    "functional-qml-application-uses-friends-mean")

        self._add_result(t, n, s, link)

    def _validate_qml_import_versions(self, frameworks, qml):
        '''Validate a QML import version against available frameworks'''
        pat_imp = re.compile(r'^\s*import (Qt[\w\.]+) ([0-9\.]+)(.*)?')
        valid = True
        for match in pat_imp.findall(str(qml)):
            min_fws = {}
            try:
                min_fws = QT_IMP_FRAMEWORKS[match[0]]
            except KeyError:
                t = 'warn'
                n = self._get_check_name('qt_qml_import_unknown')
                s = 'Unsupported Qt module import of `%s`' % (match[0])
                link = None
                self._add_result(t, n, s, link)
                continue

            min_ver = None
            for item in min_fws:
                if apt_pkg.version_compare(match[1], item[0]) <= 0:
                    min_ver = item[1]
                    break

            if min_ver is None:
                continue

            for fw in frameworks:
                if apt_pkg.version_compare(fw, min_ver) < 0:
                    valid = False
                    break
        return valid

    def check_qt_qml_import_versions(self):
        '''Check that Qt/QML applications use a new framework version.'''
        if not self.is_click:
            return

        t = 'info'
        n = self._get_check_name('qt_qml_import_versions_framework')
        s = "OK"
        link = None

        frameworks = self.manifest['framework'].split(',')

        files = []
        for i in self.qml_files:
            qml = open_file_read(i).read()
            if not self._validate_qml_import_versions(frameworks, qml):
                files.append(os.path.relpath(i, self.unpack_dir))

        for binary in self.pkg_bin_files:
            qml = subprocess.run(['strings', binary], stdout=subprocess.PIPE).stdout
            if not self._validate_qml_import_versions(frameworks, qml):
                files.append(os.path.relpath(binary, self.unpack_dir))

        if len(files) > 0:
            t = 'error'
            s = 'Files contain QML imports too new for framework used: %s' % ','.join(files)
            self._add_result(t, n, s, link)

    def check_qt_framework(self):
        '''Check that Qt/QML applications use a new framework version.'''
        if not self.is_click:
            return

        t = 'info'
        n = self._get_check_name('qt_version_framework')
        s = "OK"
        link = None

        frameworks = self.manifest['framework'].split(',')

        pat_qt = re.compile(r'Qt_([0-9]+\.[0-9]+)')
        for binary in self.pkg_bin_files:
            qt = subprocess.run(['strings', binary], stdout=subprocess.PIPE).stdout
            for match in pat_qt.findall(str(qt)):
                try:
                    min_ver = QT_MIN_FRAMEWORKS[match]
                    valid = False
                    for fw in frameworks:
                        if apt_pkg.version_compare(fw, min_ver) >= 0:
                            valid = True
                            break

                    if not valid:
                        t = 'error'
                        s = 'Linked Qt version too new for framework used: %s' \
                            % os.path.relpath(binary, self.unpack_dir)
                        self._add_result(t, n, s, link)
                except KeyError:
                    t = 'error'
                    s = 'Unsupported version of Qt used'
                    self._add_result(t, n, s, link)
